#include "core.h"
#include "varmap.h"
#include "idatastream.h"
#include "stringmanager.h"
#include "reversehash.h"
#include <string.h>

#if PARANOID_HASH_CHECK
#define STRING_TYPE SimpleString
#define HASH_STRING(s) (HashedString(s))
#else
#define STRING_TYPE HashedString
#define HASH_STRING(s) (s)
#endif

VarMap::VarMap()
:	m_Vars()
{}

VarMap::~VarMap()
{
	m_Vars.Clear();
}

void VarMap::Clear()
{
	m_Vars.Clear();
}

void VarMap::Load( const IDataStream& Stream )
{
	if( MAGIC_ID("VMAP") == Stream.ReadUInt32() )
	{
		uint NumVars = Stream.ReadUInt32();

		for( uint VarIndex = 0; VarIndex < NumVars; ++VarIndex )
		{
			ConfigVar ConfigVar;
			HashedString Key( Stream.ReadUInt32() );
			ConfigVar.m_Type = ( ConfigVar::EVarType )Stream.ReadUInt8();

			switch( ConfigVar.m_Type )
			{
			case ConfigVar::EVT_Bool:
				ConfigVar.m_Bool = ( Stream.ReadUInt8() != 0 );
				break;
			case ConfigVar::EVT_Int:
				ConfigVar.m_Int = Stream.ReadInt32();
				break;
			case ConfigVar::EVT_Float:
				ConfigVar.m_Float = Stream.ReadFloat();
				break;
			case ConfigVar::EVT_String:
				{
					uint Length = Stream.ReadUInt32();

					char* pString = new char[ Length + 1 ];
					Stream.Read( Length, pString );
					pString[ Length ] = '\0';
					StringManager::AddString( StringManager::ESL_Permanent, pString );

					ConfigVar.m_String = pString;
					ConfigVar.m_Hash = Stream.ReadUInt32();

					if( ReverseHash::IsEnabled() )
					{
						ReverseHash::RegisterHash( ConfigVar.m_Hash, ConfigVar.m_String );
					}
				}
				break;
			default:
				WARNDESC( "Bad config var type" );
				break;
			}

			SetConfigVar( Key, ConfigVar );
		}
	}
}

void VarMap::Save( const IDataStream& Stream )
{
	Stream.WriteUInt32( MAGIC_ID("VMAP") );		// Magic ID: VMAP
	Stream.WriteUInt32( m_Vars.Size() );		// Number of variables

	FOR_EACH_MAP( VarIterator, m_Vars, HashedString, ConfigVar )
	{
		const HashedString& Key = VarIterator.GetKey();
		const ConfigVar& Value = VarIterator.GetValue();

		Stream.WriteUInt32( Key.GetHash() );	// Hashed var name
		Stream.WriteUInt8( ( uint8 )Value.m_Type );

		switch( Value.m_Type )
		{
		case ConfigVar::EVT_Bool:
			Stream.WriteUInt8( Value.m_Bool );
			break;
		case ConfigVar::EVT_Int:
			Stream.WriteInt32( Value.m_Int );
			break;
		case ConfigVar::EVT_Float:
			Stream.WriteFloat( Value.m_Float );
			break;
		case ConfigVar::EVT_String:
			{
				// Doesn't write the terminating null
				uint Length = ( uint )strlen( Value.m_String );
				Stream.WriteUInt32( Length );
				Stream.Write( Length, Value.m_String );
				Stream.WriteUInt32( Value.m_Hash.GetHash() );
			}
			break;
		default:
			WARNDESC( "Bad config var type" );
			break;
		}
	}
}

ConfigVar& VarMap::GetConfigVar( const STRING_TYPE& Name )
{
	const HashedString HashedName = HASH_STRING( Name );

	Map< HashedString, ConfigVar >::Iterator VarIter = m_Vars.Search( HashedName );
	if( VarIter.IsNull() )
	{
		VarIter = m_Vars.Insert( HashedName, ConfigVar() );
	}
	return *VarIter;
}

bool VarMap::GetBool( const STRING_TYPE& Name, bool Default /*= false*/ )
{
	const ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == ConfigVar::EVT_Bool || Var.m_Type == ConfigVar::EVT_None );
#endif

	if( Var.m_Type == ConfigVar::EVT_Bool )
	{
		return Var.m_Bool;
	}
	return Default;
}

int VarMap::GetInt( const STRING_TYPE& Name, int Default /*= 0*/ )
{
	const ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == ConfigVar::EVT_Int || Var.m_Type == ConfigVar::EVT_None );
#endif

	if( Var.m_Type == ConfigVar::EVT_Int )
	{
		return Var.m_Int;
	}
	return Default;
}

float VarMap::GetFloat( const STRING_TYPE& Name, float Default /*= 0.0f*/ )
{
	const ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == ConfigVar::EVT_Float || Var.m_Type == ConfigVar::EVT_None );
#endif

	if( Var.m_Type == ConfigVar::EVT_Float )
	{
		return Var.m_Float;
	}
	return Default;
}

const char* VarMap::GetString( const STRING_TYPE& Name, const char* Default /*= NULL*/ )
{
	const ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == ConfigVar::EVT_String || Var.m_Type == ConfigVar::EVT_None );
#endif

	if( Var.m_Type == ConfigVar::EVT_String )
	{
		return Var.m_String;
	}
	return Default;
}

HashedString VarMap::GetHash( const STRING_TYPE& Name, const HashedString& Default /*= HashedString::NullString*/ )
{
	const ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == ConfigVar::EVT_String || Var.m_Type == ConfigVar::EVT_None );
#endif

	if( Var.m_Type == ConfigVar::EVT_String )
	{
		return Var.m_Hash;
	}
	return Default;
}

void VarMap::SetConfigVar( const STRING_TYPE& Name, const ConfigVar& NewVar )
{
	ConfigVar& Var = GetConfigVar( Name );

#if STRICT_TYPE_CHECK
	DEBUGASSERT( Var.m_Type == NewVar.m_Type || Var.m_Type == ConfigVar::EVT_None );
#endif

	Var = NewVar;
}

void VarMap::SetBool( const STRING_TYPE& Name, bool Value )
{
	ConfigVar Var;

	Var.m_Type = ConfigVar::EVT_Bool;
	Var.m_Bool = Value;

	SetConfigVar( Name, Var );
}

void VarMap::SetInt( const STRING_TYPE& Name, int Value )
{
	ConfigVar Var;

	Var.m_Type = ConfigVar::EVT_Int;
	Var.m_Int = Value;

	SetConfigVar( Name, Var );
}

void VarMap::SetFloat( const STRING_TYPE& Name, float Value )
{
	ConfigVar Var;

	Var.m_Type = ConfigVar::EVT_Float;
	Var.m_Float = Value;

	SetConfigVar( Name, Var );
}

void VarMap::SetString( const STRING_TYPE& Name, const char* Value )
{
	ConfigVar Var;

	Var.m_Type = ConfigVar::EVT_String;
	Var.m_String = Value;
	Var.m_Hash = Value;

	SetConfigVar( Name, Var );
}

