#ifndef LUAINTERPRETER_H
#define LUAINTERPRETER_H

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#include <typeinfo>

class LuaInterpreter
{
public:

	struct UserData
	{
		const char *name;
	};

	static LuaInterpreter	*GetInstance();
	static void				DeleteInstance();

	void PushPath( const char *Path );
	void PopPath();

	bool IsValid() { return m_LuaState != NULL; }

	int Eval( const char *Data, const char *Name = "data", bool ResetStack = false );
	bool EvalInContext( const char *Data, const char *Context, const char *Name = "data" );
	bool RunInContext( const char *Data, const char *Context, const char *Name = "data" );
	bool CallGetvarHandler( const HashedString &Var, const HashedString &Context );
	bool RunFile( const char *Filename );

	bool AddContext( const char *Context );
	bool AddCFunction( const char *Name, lua_CFunction Func, const char *Context = NULL );

	void SetNil( const char *Name, const char *Context = NULL );
	void SetString( const char *Name, const char *Value, const char *Context = NULL );
	void SetInteger( const char *Name, int Value, const char *Context = NULL );
	void SetNumber( const char *Name,  double Value, const char *Context = NULL );
	void SetBoolean( const char *Name, bool Value, const char *Context = NULL );

	const char *LastError() { return m_LastError; }
	void ClearError();

	bool IsString();
	SimpleString PopString();
	bool IsInteger();
	int32 PopInteger();
	bool IsNumber();
	double PopNumber();
	bool IsFloat() { return IsNumber(); }
	double PopFloat() { return PopNumber(); }
	bool IsBoolean();
	bool PopBoolean();
	void Discard();

	template<class T> void RegisterUsertype( const lua_CFunction Gc, const lua_CFunction Index = NULL, const luaL_Reg *Methods = NULL, const lua_CFunction NewIndex = NULL )
	{
		const char *UDName = typeid( T ).name();
		luaL_newmetatable( m_LuaState, UDName );
		if( Gc )
		{
			lua_pushcfunction( m_LuaState, Gc );
			lua_setfield( m_LuaState, -2, "__realgc" );
			lua_pushcfunction( m_LuaState, UD__Gc );
			lua_setfield( m_LuaState, -2, "__gc" );
		}
		if( Methods )
		{
			lua_newtable( m_LuaState );
			luaL_setfuncs( m_LuaState, Methods, 0 );
			lua_setfield( m_LuaState, -2, "__methods" );
		}
		if( Index )
		{
			lua_pushcfunction( m_LuaState, Index );
			lua_setfield( m_LuaState, -2, "__realindex" );
		}
		if ( Methods || Index )
		{
			lua_pushcfunction( m_LuaState, UD__Index );
			lua_setfield( m_LuaState, -2, "__index" );
		}
		if( NewIndex )
		{
			lua_pushcfunction( m_LuaState, NewIndex );
			lua_setfield( m_LuaState, -2, "__newindex" );
		}
		lua_pop( m_LuaState, 1 );
	}

	// also call new on created onject
	template<class T> static T* PushUserdata( lua_State *L )
	{
		T *Obj = PushRawUserdata<T>( L );
		new (Obj) T;
		return Obj;
	}

	template<class T, class T1> static T* PushUserdata( lua_State *L, T1 Arg )
	{
		T *Obj = PushRawUserdata<T>( L );
		new (Obj) T( Arg );
		return Obj;
	}

	template<class T> T* PushUserdata()
	{
		return PushUserdata<T>( m_LuaState );
	}

	template<class T> void SetUserdata( const char *Name, T &Value, const char *Context = NULL )
	{
		PushUserdata( Value );
		SetVar( Name, Context );
	}

	template<class T> T *PopUserdata()
	{
		const char *UDName = typeid(T).name();
		T *Obj = ( T* ) luaL_testudata( m_LuaState, -1, UDName );
		lua_pop( m_LuaState, 1 );
		return Obj;
	}

	template<class T> static T *GetUserdata( lua_State *L, int Idx )
	{
		const char *UDName = typeid(T).name();
		T *Obj = ( T* ) luaL_testudata( L, Idx, UDName );
		return Obj;
	}

	template<class T> T *GetUserdata( int Idx )
	{
		return GetUserdata<T>( m_LuaState, Idx );
	}

	void PrintStack();

	static void DumpStack( LuaInterpreter *Lua ) { DumpStack( Lua->m_LuaState ); }

	int RegisterNamespace( const char *Name, const luaL_Reg *Func );

	lua_State *GetState() { return m_LuaState; }

private:
	// no new, just the memory
	template<class T> static T* PushRawUserdata( lua_State *L )
	{
		const char *UDName = typeid( T ).name();
		T *Obj = ( T* ) lua_newuserdata( L, sizeof( T ) );
		luaL_getmetatable( L, UDName );
		lua_setmetatable( L, -2 );
		return Obj;
	}

	template<class T> T* PushRawUserdata()
	{
		return PushRawUserdata<T>( m_LuaState );
	}

	static LuaInterpreter *m_Instance;

	lua_State *m_LuaState;
	char *m_LastError;

	LuaInterpreter();
	~LuaInterpreter();

	bool Pcall( int Nargs, int Nrets );
	int Error( const char *msg, ... );
	int GetContext( const char *Context );
	int GetContextByHash( const HashedString& Context );
	int DoEvalInContext( const char *Data, const char *Context, const char *Name, int NResults );
	void SetVar( const char *Name, const char *Context );

	static void DumpStack( lua_State *L );
	static int UD__Index( lua_State *L );
	static int UD__Gc( lua_State *L );
};

#endif // LUAINTERPRETER_H
