#include "core.h"
#include "simplestring.h"
#include "hashedstring.h"
#include "luainterpreter.h"
#include "lualibraries.h"
#include "file.h"
#include "packstream.h"
#include "modmanager.h"

static int FsList( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	Array< SimpleString > Contents;

	if( !FileUtil::IsDir( Pathname ) )
	{
		return luaL_error( L, "'%s' is not a directory", Pathname );
	}

	FileUtil::GetAllEntries( Pathname, Contents );

	lua_newtable( L );
	int cnt = 1;
	for( uint i = 0; i < Contents.Size(); ++i )
	{
		lua_pushstring( L, Contents[i].CStr() );
		lua_rawseti( L, -2, cnt++ );
	}

	return 1;
}

static int FsIsFile( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	lua_pushboolean( L, FileUtil::IsFile( Pathname ) );
	return 1;
}

static int FsIsDir( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	lua_pushboolean( L, FileUtil::IsDir( Pathname ) );
	return 1;
}

static int FsGetExecDir( lua_State *L )
{
	SimpleString ExecDir = FileUtil::GetExecutableDirectory();
	lua_pushstring( L, ExecDir.CStr() );
	return 1;
}

static int FsGetExecName( lua_State *L )
{
	SimpleString ExecName = FileUtil::GetFullExecutableName();
	lua_pushstring( L, FileUtil::StripLeadingFolders( ExecName.CStr() ) );
	return 1;
}

static int FsGetModDir( lua_State *L )
{
	SimpleString ModPath = ModManager::GetInstance()->GetCurrentModPath();
	if( ModPath == "" )
	{
		ModPath = FileUtil::GetWorkingDirectory();
	}
	lua_pushstring( L, ModPath.CStr() );
	return 1;
}

static int FsGetCurrentDir( lua_State *L )
{
	SimpleString CurrentDir = FileUtil::GetWorkingDirectory();
	lua_pushstring( L, CurrentDir.CStr() );
	return 1;
}

static int FsSetCurrentDir( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	FileUtil::ChangeWorkingDirectory( Pathname );
	return 0;
}

static int FsBasename( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	SimpleString Basename = FileUtil::StripExtension( FileUtil::StripLeadingFolders( Pathname ) );
	lua_pushstring( L, Basename.CStr() );
	return 1;
}

static int FsFilename( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	lua_pushstring( L, FileUtil::StripLeadingFolders( Pathname ) );
	return 1;
}

static int FsDirname( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	SimpleString Dirname = FileUtil::StripFilename( Pathname );
	lua_pushstring( L, Dirname.CStr() );
	return 1;
}

static int FsExtension( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	lua_pushstring( L, FileUtil::StripAllButExtension( Pathname ) );
	return 1;
}

static int FsNormalize( lua_State *L )
{
	const char *Pathname = luaL_checkstring( L, 1 );
	lua_pushstring( L, FileUtil::Normalize( Pathname ).CStr() );
	return 1;
}

/* Function list
 */
static const luaL_Reg Functions[] = {
	{ "list", FsList },
	{ "isfile", FsIsFile },
	{ "isdir", FsIsDir },
	{ "getexecdir", FsGetExecDir },
	{ "getexecname", FsGetExecName },
	{ "getmoddir", FsGetModDir },
	{ "getcurrentdir", FsGetCurrentDir },
	{ "setcurrentdir", FsSetCurrentDir },
	{ "basename", FsBasename },
	{ "filename", FsFilename },
	{ "dirname", FsDirname },
	{ "extension", FsExtension },
	{ "normalize", FsNormalize },
 	{ NULL, NULL }
};

int LuaLibraries::FsInit()
{
	LuaInterpreter *Lua = LuaInterpreter::GetInstance();
	return Lua->RegisterNamespace( "fs", Functions );
}
