#include "core.h"
#include "hashedstring.h"
#include "luainterpreter.h"
#include "lualibraries.h"
#include "eldritchroom.h"
#include "eldritchworld.h"
#include "eldritchworldgen.h"
#include "dynamicmemorystream.h"
#include "packstream.h"
#include "filestream.h"
#include "modmanager.h"
#include "configmanager.h"
#include "runtimeerror.h"

// from eldritchtools.cpp
static const vidx_t	kInvalidVoxelIndex	= 0xffffffff;
static const vval_t	kInvalidVoxelValue	= 0xff;

class EldritchRoomBuilder : public EldritchRoom
{
public:
	EldritchRoomBuilder()
	: m_RoomSizeX( 0 )
	, m_RoomSizeY( 0 )
	, m_RoomSizeZ( 0 )
	, m_NumVoxels( 0 )
	{
		// default values are from current config. In the context of eldritch,
		// these should not be changed anyway as havoc will ensue if you do!
		m_RoomSizeX = ConfigManager::GetInt( "RoomSizeX", 12, "EldritchWorld" );
		m_RoomSizeY = ConfigManager::GetInt( "RoomSizeY", 12, "EldritchWorld" );
		m_RoomSizeZ = ConfigManager::GetInt( "RoomSizeZ", 8, "EldritchWorld" );
		m_RoomScalarX = 1;
		m_RoomScalarY = 1;
		m_RoomScalarZ = 1;
		m_NumVoxels = m_RoomSizeX * m_RoomSizeY * m_RoomSizeZ;
		m_VoxelMap.ResizeZero( m_NumVoxels );
	}

	~EldritchRoomBuilder() {};

	void SetVoxel( uint X, uint Y, uint Z, vval_t Voxel )
	{
		uint Index = GetVoxelIndexForCoords( X, Y, Z );
		if( Index != kInvalidVoxelIndex )
		{
			m_VoxelMap[ Index ] = Voxel;
		}
	}

	vval_t GetVoxel( uint X, uint Y, uint Z )
	{
		uint Index = GetVoxelIndexForCoords( X, Y, Z );
		if( Index != kInvalidVoxelIndex )
		{
			return m_VoxelMap[ Index ];
		}
		return kInvalidVoxelValue;
	}

	void SetSpawner( uint X, uint Y, uint Z, SimpleString Name, uint8 Orientation )
	{
		uint Index = GetVoxelIndexForCoords( X, Y, Z );
		if( Index != kInvalidVoxelIndex )
		{
			SSpawner Spawner;
			Spawner.m_SpawnerDef = Name;
			Spawner.m_Location = Index;
			Spawner.m_Orientation = Orientation;

			uint ExistingSpawner = FindSpawner( Index );
			if( ExistingSpawner != kInvalidVoxelIndex )
			{
				m_Spawners[ ExistingSpawner ] = Spawner;
			}
			else
			{
				m_Spawners.PushBack( Spawner );
			}
		}
	}

	SSpawner const *GetSpawner( uint X, uint Y, uint Z )
	{
		uint Index = GetVoxelIndexForCoords( X, Y, Z );
		if( Index != kInvalidVoxelIndex )
		{
			uint ExistingSpawner = FindSpawner( Index );
			if( ExistingSpawner != kInvalidVoxelIndex )
			{
				return &m_Spawners[ ExistingSpawner ];
			}
		}
		return NULL;
	}

	SSpawner const *GetSpawner( uint Index )
	{
		if( Index < m_Spawners.Size() )
		{
			return &m_Spawners[ Index ];
		}
		return NULL;
	}

	bool IsValidCoord( uint X, uint Y, uint Z )
	{
		return GetVoxelIndexForCoords( X, Y, Z ) != kInvalidVoxelIndex;
	}

	void GetRoomSize( uint &X, uint &Y, uint &Z )
	{
		X = m_RoomSizeX;
		Y = m_RoomSizeY;
		Z = m_RoomSizeZ;
	}

	HashedString GetUsedWorldDef()
	{
		return m_UsedWorldDef;
	}

	void SetUsedWorldDef( HashedString &UsedWorldDef )
	{
		m_UsedWorldDef = UsedWorldDef;
	}

	// from eldritchtools
	bool GetCoordsForVoxel( const vidx_t VoxelIndex, uint& X, uint& Y, uint& Z )
	{
		if( VoxelIndex < m_NumVoxels )
		{
			X = VoxelIndex % m_RoomSizeX;
			Y = ( VoxelIndex / m_RoomSizeX ) % m_RoomSizeY;
			Z = ( ( VoxelIndex / m_RoomSizeX ) / m_RoomSizeY ) % m_RoomSizeZ;
			return true;
		}
		return false;
	}

private:
	// from eldritchtools
	uint GetVoxelIndexForCoords( uint X, uint Y, uint Z )
	{
		if( X < m_RoomSizeX && Y < m_RoomSizeY && Z < m_RoomSizeZ )
		{
			return X + ( Y * m_RoomSizeX ) + ( Z * m_RoomSizeX * m_RoomSizeY );
		}
		return kInvalidVoxelIndex;
	}

	uint FindSpawner( uint VoxelIndex )
	{
		for( uint i = 0; i < m_Spawners.Size(); ++i )
		{
			if( m_Spawners[i].m_Location == VoxelIndex )
			{
				return i;
			}
		}
		return kInvalidVoxelIndex;
	}

	uint m_RoomSizeX;
	uint m_RoomSizeY;
	uint m_RoomSizeZ;
	uint m_NumVoxels;
};

static int RoomCreate( lua_State *L )
{
	LuaInterpreter::PushUserdata<EldritchRoomBuilder>( L );
	return 1;
}

static int RoomLoad( lua_State *L )
{
	const char *Filename = luaL_checkstring( L, 1 );
	EldritchRoomBuilder *Room = LuaInterpreter::PushUserdata<EldritchRoomBuilder>( L );
	int Top = lua_gettop( L );
	try
	{
		PackStream Stream = ModManager::OpenStream( Filename );
		// if the room was loaded from a memory stream, that will still be present on  the stack
		lua_settop( L, Top );
		if( !Stream.EOS() )
		{
			Room->Load( Stream );
		}
		else
		{
			lua_pop( L, 1 );
			lua_pushnil( L );
		}
	}
	catch (RuntimeError Err)
	{
		lua_pop( L, 1 );
		lua_pushnil( L );
	}
	return 1;
}

/* Function list
 */
static const luaL_Reg Functions[] = {
	{ "create", RoomCreate },
	{ "load", RoomLoad },
 	{ NULL, NULL }
};

static int RoomSave( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	const char *Filename = luaL_checkstring( L, 2 );
	Room->Save( FileStream( Filename, FileStream::EFM_Write ) );
	return 0;
}

static int RoomSaveTo( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	DynamicMemoryStream *Stream = LuaInterpreter::GetUserdata<DynamicMemoryStream>( L, 2 );
	if( Room && Stream )
	{
		Room->Save( *Stream );
	}
	return 0;
}

static int RoomSetVoxel( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	uint X = luaL_checkinteger( L, 2 );
	uint Y = luaL_checkinteger( L, 3 );
	uint Z = luaL_checkinteger( L, 4 );
	uint8 Voxel = luaL_checkinteger( L, 5 ) & 0xFF;
	Room->SetVoxel( X, Y, Z, Voxel );
	return 0;
}

static int RoomGetVoxel( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	uint X = luaL_checkinteger( L, 2 );
	uint Y = luaL_checkinteger( L, 3 );
	uint Z = luaL_checkinteger( L, 4 );
	uint8 Voxel = Room->GetVoxel( X, Y, Z );
	if( Voxel == kInvalidVoxelValue )
	{
		lua_pushnil( L );
	}
	else
	{
		lua_pushinteger( L, Voxel );
	}
	return 1;
}

static int OrientationFromName( const char *Direction )
{
	switch( *Direction )
	{
	case 'w': case 'W': return ESO_West;
	case 'e': case 'E': return ESO_East;
	case 's': case 'S': return ESO_South;
	default: ;
	}
	return ESO_North;
}

static uint ToOrientation( lua_State *L, int Pos )
{
	if( lua_gettop( L ) < Pos )
	{
		return ESO_North;
	}
	if( lua_type( L, Pos ) == LUA_TSTRING )
	{
		const char *Direction = lua_tostring( L, Pos );
		return OrientationFromName( Direction );
	}
	uint Direction = luaL_checkinteger( L, Pos );
	return Direction > ESO_East ? ESO_North : Direction;
}

static int PushOrientationName( lua_State *L, uint8 Orientation )
{
	Orientation = Orientation > ESO_East ? ESO_North : Orientation;
	switch( Orientation )
	{
	case ESO_North:
		lua_pushstring( L, "n" );
		break;
	case ESO_West:
		lua_pushstring( L, "w" );
		break;
	case ESO_South:
		lua_pushstring( L, "s" );
		break;
	case ESO_East:
		lua_pushstring( L, "e" );
		break;
	}
	return 1;
}

static int RoomSetSpawner( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	uint X = luaL_checkinteger( L, 2 );
	uint Y = luaL_checkinteger( L, 3 );
	uint Z = luaL_checkinteger( L, 4 );
	const char *Spawner = luaL_checkstring( L, 5 );
	uint Orientation = ToOrientation( L, 6 );
	Room->SetSpawner( X, Y, Z, Spawner, Orientation );
	return 0;
}

static int RoomGetSpawner( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	const EldritchRoom::SSpawner *Spawner = NULL;
	uint X = 0, Y = 0, Z = 0;

	if( lua_gettop( L ) == 2 )
	{
		uint Index = luaL_checkinteger( L, 2 );
		Spawner = Room->GetSpawner( Index );
	}
	else
	{
		X = luaL_checkinteger( L, 2 );
		Y = luaL_checkinteger( L, 3 );
		Z = luaL_checkinteger( L, 4 );
		Spawner = Room->GetSpawner( X, Y, Z );
	}
	if( !Spawner )
	{
		return 0;
	}
	lua_pushstring( L, Spawner->m_SpawnerDef.CStr() );
	PushOrientationName( L, Spawner->m_Orientation );
	Room->GetCoordsForVoxel( Spawner->m_Location, X, Y, Z );
	lua_pushinteger( L, X );
	lua_pushinteger( L, Y );
	lua_pushinteger( L, Z );
	return 5;
}

static int RoomGetWorldDef( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	HashedString WorldDef = Room->GetUsedWorldDef();
	lua_pushinteger( L, WorldDef.GetHash() );
	return 1;
}

static int RoomSetWorldDef( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	HashedString Hash;
	if( lua_type( L, 2 ) == LUA_TSTRING )
	{
		const char *Name = lua_tostring( L, 2 );
		Hash = HashedString( Name );
	}
	else
	{
		Hash = luaL_checkinteger( L, 2 );
	}
	Room->SetUsedWorldDef( Hash );
	return 0;
}

static const luaL_Reg Methods[] = {
	{ "save", RoomSave },
	{ "saveto", RoomSaveTo },
	{ "setvoxel", RoomSetVoxel },
	{ "getvoxel", RoomGetVoxel },
	{ "setspawner", RoomSetSpawner },
	{ "getspawner", RoomGetSpawner },
	{ "getworlddef", RoomGetWorldDef },
	{ "setworlddef", RoomSetWorldDef },
 	{ NULL, NULL }
};

static int Gc( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	Room->~EldritchRoomBuilder();
	return 0;
}

static int Index( lua_State *L )
{
	EldritchRoomBuilder *Room = LuaInterpreter::GetUserdata<EldritchRoomBuilder>( L, 1 );
	const char *What = lua_tostring( L, 2 );
	if ( !strcmp( What, "sizex") )
	{
		uint X, Y, Z;
		Room->GetRoomSize( X, Y, Z );
		lua_pushinteger( L, X );
	}
	else if ( !strcmp( What, "sizey" ) )
	{
		uint X, Y, Z;
		Room->GetRoomSize( X, Y, Z );
		lua_pushinteger( L, Y );
	}
	else if ( !strcmp( What, "sizez" ) )
	{
		uint X, Y, Z;
		Room->GetRoomSize( X, Y, Z );
		lua_pushinteger( L, Z );
	}
	else
	{
		lua_pushnil( L );
	}
	return 1;
}

int LuaLibraries::RoomInit()
{
	LuaInterpreter *Lua = LuaInterpreter::GetInstance();
	Lua->RegisterUsertype<EldritchRoomBuilder>( Gc, Index, Methods );
	return Lua->RegisterNamespace( "room", Functions );
}
