#include "core.h"
#include "eldritchworldgen.h"
#include "configmanager.h"
#include "eldritchframework.h"
#include "eldritchworld.h"
#include "eldritchroom.h"
#include "packstream.h"
#include "modmanager.h"
#include "mathfunc.h"
#include "reversehash.h"
#include "filestream.h"
#include "wbworld.h"
#include "wbentity.h"
#include "Components/wbcompeldtransform.h"
#include "Components/wbcompeldcollision.h"
#include "mathcore.h"
#include "matrix.h"
#include "wbeventmanager.h"
#include "collisioninfo.h"
#include "file.h"
#include "eldritchgame.h"
#include "eldritchpreviewworldgen.h"

enum ERoomId
{
	FILL,
	CORNER,
	SIDE,
	UP,
	DOWN,
	DEFAULT
};

struct SRoomPlacement
{
	int X, Y, Z;
	ERoomId RoomId;
	int Xforms;
};

#define NONE EldritchWorldGen::ERT_None
#define M_X EldritchWorldGen::ERT_MirrorX
#define M_Y EldritchWorldGen::ERT_MirrorY
#define ROT EldritchWorldGen::ERT_Rotate90

static const int NumRoomPlacements = 11;
static const SRoomPlacement RoomPlacements[ NumRoomPlacements ] =
{
	{ 1, 1, 2, DOWN,	NONE },
	{ 0, 0, 1, CORNER,	NONE },
	{ 1, 0, 1, SIDE,	NONE },
	{ 2, 0, 1, CORNER,	M_X },
	{ 0, 1, 1, SIDE,	ROT | M_Y },
	{ 1, 1, 1, DEFAULT,	NONE },
	{ 2, 1, 1, SIDE,	ROT },
	{ 0, 2, 1, CORNER,	M_Y },
	{ 1, 2, 1, SIDE,	M_Y },
	{ 2, 2, 1, CORNER,	M_X | M_Y },
	{ 1, 1, 0, UP,		NONE },
};

EldritchPreviewWorldGen::EldritchPreviewWorldGen()
:	m_UpStartRoomName( "" )
,	m_DownStartRoomName( "" )
,	m_LevelStartRoomName( "" )
{
}

EldritchPreviewWorldGen::~EldritchPreviewWorldGen()
{
}

void EldritchPreviewWorldGen::Initialize( const SimpleString& WorldGenDefinitionName )
{
	EldritchWorldGen::Initialize( WorldGenDefinitionName );

	MAKEHASH( WorldGenDefinitionName );

	m_RoomFillName = ConfigManager::GetString( "RoomFill", "", sWorldGenDefinitionName );
	ASSERT( m_RoomFillName != "" );
	m_RoomCornerName = ConfigManager::GetString( "RoomCorner", "", sWorldGenDefinitionName );
	ASSERT( m_RoomCornerName != "" );
	m_RoomSideName = ConfigManager::GetString( "RoomSide", "", sWorldGenDefinitionName );
	ASSERT( m_RoomSideName != "" );
	m_RoomUpName = ConfigManager::GetString( "RoomUp", "", sWorldGenDefinitionName );
	ASSERT( m_RoomUpName != "" );
	m_RoomDownName = ConfigManager::GetString( "RoomDown", "", sWorldGenDefinitionName );
	ASSERT( m_RoomDownName != "" );
	m_RoomDefaultName = ConfigManager::GetString( "RoomDefault", "", sWorldGenDefinitionName );
	ASSERT( m_RoomDefaultName != "" );
	m_UpStartRoomName = ConfigManager::GetString( "RoomStartUp", "", sWorldGenDefinitionName );
	ASSERT( m_UpStartRoomName != "" );
	m_DownStartRoomName = ConfigManager::GetString( "RoomStartDown", "", sWorldGenDefinitionName );
	ASSERT( m_DownStartRoomName != "" );
	m_LevelStartRoomName = ConfigManager::GetString( "RoomStartLevel", "", sWorldGenDefinitionName );
	ASSERT( m_LevelStartRoomName != "" );
}

const SimpleString& EldritchPreviewWorldGen::GetRoomName( int RoomId )
{
	switch ((ERoomId) RoomId)
	{
		case CORNER: return m_RoomCornerName;
		case SIDE: return m_RoomSideName;
		case UP: return m_RoomUpName;
		case DOWN: return m_RoomDownName;
		case DEFAULT: return m_RoomDefaultName;
		case FILL:
		default: return m_RoomFillName;
	}
}

void EldritchPreviewWorldGen::FillWorld()
{
	int MapSizeX = m_World->GetMapSizeX();
	int MapSizeY = m_World->GetMapSizeY();
	int MapSizeZ = m_World->GetMapSizeZ();

	int X, Y, Z;
	SimpleString FillerRoom = GetRoomName( FILL );

	for ( X = 0; X < MapSizeX; ++X )
	{
		for ( Y = 0; Y < MapSizeY; ++Y )
		{
			for ( Z = 0; Z < MapSizeZ; ++Z )
			{
				int RoomIndex = GetRoomIndex( X, Y, Z );
				SGenRoom& GenRoom = m_GeneratedRooms[ RoomIndex ];
				GenRoom.m_UsePrescribedRoom = true;
				GenRoom.m_Exits = ERE_None;
				GenRoom.m_PrescribedFilename = FillerRoom;
				GenRoom.m_PrescribedTransform = ERT_None;
				m_Maze_UnopenRoomSet.Remove( RoomIndex );
				m_Maze_LockedRoomSet.Insert( RoomIndex );
			}
		}
	}
}

void EldritchPreviewWorldGen::Generate( const SimpleString& TestRoomName )
{
	PROFILE_FUNCTION;

	ERoomExits Exits = ERE_None;
	GetExitsFromFilename( TestRoomName, Exits );

	TPersistence& Persistence = EldritchGame::StaticGetTravelPersistence();
	STATIC_HASHED_STRING( Hard );
	m_IsHardMode = Persistence.GetBool( sHard );

	InitializeMazeGen();

	FillWorld();
	
	// add base rooms
	for ( int RoomNo = 0; RoomNo < NumRoomPlacements; ++RoomNo )
	{
		const SRoomPlacement& Placement = RoomPlacements[ RoomNo ];
		int RoomIndex = GetRoomIndex( Placement.X, Placement.Y, Placement.Z );
		SGenRoom& GenRoom = m_GeneratedRooms[ RoomIndex ];
		GenRoom.m_UsePrescribedRoom = true;
		GenRoom.m_Exits = ERE_None;
		GenRoom.m_PrescribedFilename = GetRoomName( Placement.RoomId );
		GenRoom.m_PrescribedTransform = (ERoomXform) Placement.Xforms;
	}

	// add tested room
	int RoomIndex = GetRoomIndex( 1, 1, 1 );
	SGenRoom& GenRoom = m_GeneratedRooms[ RoomIndex ];
	GenRoom.m_UsePrescribedRoom = true;
	GenRoom.m_Exits = Exits;
	GenRoom.m_PrescribedFilename = TestRoomName;
	GenRoom.m_PrescribedTransform = ERT_None;

	if ( !RoomHasPlayerSpawner( TestRoomName ) )
	{
		// Default: Starting point is to the south, facing north
		SimpleString StartRoomName = m_LevelStartRoomName;
		int StartIndex = GetRoomIndex( 1, 0, 1 );
		int StartXforms = ERT_None;
		if ( Exits & ERE_North )
		{
			StartIndex = GetRoomIndex( 1, 2, 1 );
			StartXforms = ERT_MirrorY;
		}
		else if ( Exits & ERE_West )
		{
			StartIndex = GetRoomIndex( 0, 1, 1 );
			StartXforms = ERT_Rotate90 | ERT_MirrorY;
		}
		else if ( Exits & ERE_East )
		{
			StartIndex = GetRoomIndex( 2, 1, 1 );
			StartXforms = ERT_Rotate90;
		}
		else if ( Exits & ERE_Down )
		{
			StartRoomName = m_UpStartRoomName;
			StartIndex = GetRoomIndex( 1, 1, 0 );
		}
		else if ( Exits & ERE_Up )
		{
			StartRoomName = m_DownStartRoomName;
			StartIndex = GetRoomIndex( 1, 1, 2 );
		}
		
		ERoomExits StartExits = ERE_None;
		GetExitsFromFilename( StartRoomName, StartExits );

		SGenRoom& StartRoom = m_GeneratedRooms[ StartIndex ];
		StartRoom.m_UsePrescribedRoom = true;
		StartRoom.m_Exits = StartExits;
		StartRoom.m_PrescribedFilename = StartRoomName;
		StartRoom.m_PrescribedTransform = (ERoomXform) StartXforms;
	}

	PopulateWorld();
}

bool EldritchPreviewWorldGen::RoomHasPlayerSpawner( const SimpleString RoomFilename )
{
	EldritchRoom Room;
	Room.Load( ModManager::OpenStream( RoomFilename.CStr() ) );
	const Array<EldritchRoom::SSpawner>& Spawners = Room.GetSpawners();
	const uint NumSpawners = Spawners.Size();
	for( uint SpawnerIndex = 0; SpawnerIndex < NumSpawners; ++SpawnerIndex )
	{
		const EldritchRoom::SSpawner&	Spawner			= Spawners[ SpawnerIndex ];
		const SimpleString&				RoomSpawnerDef	= Spawner.m_SpawnerDef;
		const SimpleString				SpawnerDef		= TryGetHardSpawner( m_World->GetSpawnerOverride( RoomSpawnerDef ) );

		MAKEHASH( SpawnerDef );

		const SimpleString SubSpawnerDef = GetSubSpawnerDef( SpawnerDef );

		if( SubSpawnerDef == "" )
		{
			// It's valid for a subspawner to be empty
			continue;
		}

		MAKEHASH( SubSpawnerDef );
		STATICHASH( Entity );
		STATICHASH( Extends );

		const SimpleString Entity = ConfigManager::GetString( sEntity, "", sSubSpawnerDef );
		const SimpleString Extends = ConfigManager::GetString( sExtends, "", Entity );

		if ( Entity == "Player" || Extends == "Player" )
		{
			return true;
		}
	}
	return false;
}